'''
description: MVFunction.py
author: Ao Wang
date: 09/17/20
'''

from IPython.display import Math, display
from sympy.abc import x, y, z

from axiomathbf.environment import isnotebook

from .gradient import Gradient
from .parametric_lines import ParametricLine
from .plane import Plane


class MVFunction(Gradient):
    def __init__(self, function, point):
        '''MVFunction class that gets tangent plane, normal line, and linearization of a function

        :param function: the multivariable function
        :type function: sympy.core.add.Add
        :param point: a point
        :type point: list of int
        :param value: value of function at a point
        :type value: int
        '''
        super().__init__(function)
        self.point = point
        p1, p2, p3 = self.point
        self.value = self.function.subs(([x, p1], [y, p2], [z, p3]))

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '{} at {} is {}'.format(self.function, self.point, self.value)

    def get_point(self):
        '''Gets the point class attribute

        :return: the point class attribute
        :rtype: list of ints
        '''
        return self.point

    def set_point(self, point):
        '''Sets the point class attribute

        :param point: the point class attribute
        :type point: list of ints
        '''
        self.point = point

    def get_value(self):
        '''Gets the value class attribute

        :return: the value class attribute
        :rtype: int
        '''
        return self.value

    def get_linearization(self):
        '''Calculates the linearization equation for local-linear approximation.

        :return: the linearization equation 
        :rtype: sympy.core.add.Add
        '''
        p1, p2, p3 = self.point
        g1, g2, g3 = self.at(self.point)
        function_at_point = self.value
        return function_at_point + g1 * (x - p1) + g2 * (y - p2) + g3 * (z - p3)

    def get_tangent_plane(self):
        '''Calculates the tangent plane of a function at a point

        :return: the tangent plane at a point at a function
        :rtype: axiomathbf.plane.Plane
        '''
        return Plane(p1=self.point, normal_vector=self.at(self.point))

    def get_normal_line(self):
        '''Calculates the normal line of a function at a point

        :return: the normal line of a function
        :rtype: axiomathbf.parametric_line.ParametricLine
        '''
        return ParametricLine(point=self.point, vector=self.at(self.point))



'''
description: directional_derivative.py
author: Ao Wang
date: 09/18/20
'''

from IPython.display import Math, display
from sympy import Matrix

from axiomathbf.environment import isnotebook

from .gradient import Gradient


class DirectionalDerivative(Gradient):
    '''The Directional Deriviative class calculates the its value, provide more info,
    like its maximum/minimum value and in which direction it increases most rapidly.

    :param function: the multivariable function
    :type function: sympy.core.add.Add
    :param point: the point at a function 
    :type point: list of int
    :param unit_vector: the directional vector
    :type unit_vector: list of int
    '''

    def __init__(self, function, point, unit_vector=None):
        super().__init__(function)
        self.point = point
        if unit_vector != None:
            unit_vector = Matrix(unit_vector)
            self.unit_vector = unit_vector/unit_vector.norm()
            self.value = self.__get_directional_diff()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return 'Function: {}\nPoint: {}\n'.format(self.function, self.point)

    def get_unit_vector(self):
        '''Gets the unit vector class attribute

        :return: the unit vector class attribute
        :rtype: the list of ints
        '''
        return self.unit_vector

    def set_unit_vector(self, unit_vector):
        '''Sets the unit vector class attribute

        :param unit_vector: the unit vector class attribute
        :type unit_vector: the list of ints
        '''
        self.unit_vector = unit_vector

    def get_point(self):
        '''Gets the point class attribute

        :return: the point class attribute
        :rtype: the list of ints
        '''
        return self.point

    def set_point(self, point):
        '''Sets the point class attribute

        :param point: the point class attribute
        :type point: the list of ints
        '''
        self.point = point

    def __get_directional_diff(self):
        '''Calculates the directional derivative at a point.

        :return: the directional derivative at a point
        :rtype: sympy.core.numbers.Integer
        '''
        return self.at(self.point).dot(self.unit_vector)

    def info(self, increasing=True):
        '''Provides more info about the directional derivative 

        :param increasing: whether the function is increasing or decreasing
        :type increasing: bool
        :return: the maximum value of directional derivative and direction vector
        :rtype: tuple of int and list
        '''
        gradient = self.at(self.point)
        maximum = gradient.norm() if increasing else -gradient.norm()
        gradient /= gradient.norm()
        unitVector = gradient if increasing else -gradient
        return (maximum, list(unitVector))



'''
description: plane class
author: ao wang
date: 09/06/2020
'''

import sympy
from IPython.display import Math, display
from sympy import Matrix, asin, sqrt
from sympy.abc import t, x, y, z

from axiomathbf.environment import isnotebook

from .parametric_lines import ParametricLine


class Plane():
    '''The plane class that can initalize with 3 points, a point and a normal vector, or equation.
    The class can also compare planes and lines to test if they are parallel, perpendicular, or skew,
    find points of intersection, and angle and distance between two 3D objects.

    :param p1: the first point
    :type p1: sympy.geometry.point.Point3D
    :param p2: the second point
    :type p2: sympy.geometry.point.Point3D
    :param p3: the third point
    :type p3: sympy.geometry.point.Point3D
    :param normal_vector: the normal vector
    :type normal_vector: list of ints
    :param eq: the plane equation
    :type eq: sympy.core.add.Add
    :param plane: the sympy plane object
    :type plane: sympy.geometry.plane.Plane
    '''

    def __init__(self, p1=None, p2=None, p3=None, normal_vector=None, eq=None):
        if p1 and p2 and p3:
            plane = sympy.Plane(p1, p2, p3)
        elif p1 and normal_vector:
            plane = sympy.Plane(p1, normal_vector=normal_vector)
        elif eq:
            # gets all the coeffients of each variable
            coeff_dict = eq.as_coefficients_dict()
            point_eq = sympy.solve(eq, x, y, z)[0]  # solves x, y, z to 0

            # Finds a point on the plane
            point = sympy.Point(
                [point.subs([(x, 0), (y, 0), (z, 0)]) for point in point_eq])
            plane = sympy.Plane(point, normal_vector=[coeff_dict.get(
                x, 0), coeff_dict.get(y, 0), coeff_dict.get(z, 0)])

        gcd = sympy.gcd(plane.normal_vector)
        plane = sympy.Plane(
            p1=plane.p1, normal_vector=[elem/gcd for elem in plane.normal_vector])
        self.plane = plane

    def __repr__(self):
        if isnotebook():
            display(Math(sympy.latex(self.plane.equation())))
            return ''
        return self.__str__()

    def __str__(self):
        return str(self.plane.equation())

    def __eq__(self, other):
        return self.plane.equation().equals(other.plane.equation())

    def get_plane(self):
        '''Gets the plane class attribute

        :return: the plane class attribute
        :rtype: sympy.geometry.plane.Plane
        '''
        return self.plane

    def set_plane(self, plane):
        '''Sets the plane class attribute

        :param plane: the sympy plane object
        :type plane: sympy.geometry.plane.Plane
        '''
        self.plane = plane

    def angle(self, other):
        '''Calculates the angle between 3D objects (Plane, ParametricLine)

        :param other: another 3D object
        :type other: Plane or ParametricLine
        :return: the angle between two 3D objects
        :rtype: sympy.core.numbers.Floats
        '''
        if isinstance(other, Plane):
            return self.plane.angle_between(other.plane)
        elif isinstance(other, ParametricLine):
            norm_vect, line_dir = Matrix(
                self.plane.normal_vector), other.vector
            return abs(asin(line_dir.dot(norm_vect)/(line_dir.norm()*norm_vect.norm())))

    def compare(self, other):
        '''Finds whether a Plane or ParametricLine are parallel, perpendicular, or neither

        :param other: another 3D object
        :type other: Plane or ParametricLine
        :return: whether the objects are perpendicular, parallel, or neither
        :rtype: str
        '''
        if isinstance(other, Plane):
            if other.plane.is_perpendicular(self.plane):
                return 'Perpendicular'
            elif other.plane.is_parallel(self.plane):
                return 'Parallel'
            else:
                return 'Neither parallel nor perpendicular'
        elif isinstance(other, ParametricLine):
            if sympy.Matrix(self.plane.normal_vector).dot(other.vector) == 0:
                return 'Parallel'
            elif (sympy.Matrix(self.plane.normal_vector).cross(other.vector)).norm() == 0:
                return 'Perpendicular'
            else:
                return 'Neither parallel nor perpendicular'

    def distance(self, other):
        '''Returns the distance between Planes, Lines, and Points

        :param other: another 3D object
        :type other: Plane, ParametricLine, or sympy.geometry.point.Point3D
        :return: the distance between 3D objects
        :rtype: sympy.core.numbers.Floats
        '''
        if isinstance(other, Plane):
            return self.plane.distance(other.plane)
        elif isinstance(other, sympy.Point):
            return self.plane.distance(other)
        elif isinstance(other, ParametricLine):
            pq, norm_vect = Matrix(
                other.point-self.plane.p1), Matrix(self.plane.normal_vector)
            return abs(pq.dot(norm_vect))/norm_vect.norm()

    def intersect(self, other):
        '''Calculates where the line or plane intersects

        :param other: another 3D object
        :type other: Plane or ParametricLine
        :return: returns the point intersection of plane and line or paramatric line of intersection betweeen two planes
        :rtype: sympy.geometry.point.Point3D or ParametricLine
        '''
        if isinstance(other, Plane):
            line = other.plane.intersection(self.plane)[0]
            return ParametricLine(point=line.p1, vector=list(line.p1-line.p2))
        elif isinstance(other, ParametricLine):
            line_eq = [
                pt + vec*t for (pt, vec) in zip(other.get_point(), other.get_vector())]
            x_line, y_line, z_line = line_eq
            t_intersect = sympy.solve(
                self.plane.equation(x_line, y_line, z_line), t)
            if t_intersect:
                return sympy.Point([elem.subs(t, t_intersect[0]) for elem in line_eq])


if __name__ == '__main__':
    x, y, z = sympy.symbols('x y z')
    # p1 = Plane(eq=5*x-3*y+4*z+1)
    # p2 = Plane(eq=2*x-2*y-4*z-9)
    # print(p1.plane.equation())
    # print(p2.plane.equation())
    # p3 = Plane(eq=x+y+z-7)
    # p4 = Plane(eq=2*x+4*z-6)
    # print(p3.intersect(p4))
    plane = Plane(eq=2*x+y-4*z-4)
    line = ParametricLine(point=[0, 2, 0], vector=[1, 3, 1])
    print(plane.intersect(line))

    p1 = Plane(eq=x-5*y+3*z-11)
    p2 = Plane(eq=-3*x+2*y-2*z+7)
    print(p1.intersect(p2))



from axiomathbf import VectorFunction, ParametricLine
import pytest
from sympy import Symbol, sqrt, ln, E, cos, sin, tan, sec, Union, Interval, oo, Reals, log, pi
from sympy.abc import t


def test_str():
    r1 = VectorFunction([3*t, t**2, t**3+1])
    r2 = VectorFunction([cos(t), sin(t), t])
    r3 = VectorFunction([ln(t), 1, t**2])
    sols = ['<3*t, t**2, t**3 + 1>',
            '<cos(t), sin(t), t>',
            '<log(t), 1, t**2>']
    for r, sol in zip((r1, r2, r3), sols):
        assert str(r) == sol


def test_eq():
    r1 = VectorFunction([6*t, 1/t, 0])
    r2 = VectorFunction([sin(t), t**3, 1-t])
    r3 = VectorFunction([2*t*cos(t), t**2, t-1])
    sols = [VectorFunction([6*t, 1/t, 0]),
            VectorFunction([sin(t), t**3, 1-t]),
            VectorFunction([2*t*cos(t), t**2, t-1])]
    for r, sol in zip((r1, r2, r3), sols):
        assert r == sol


def test_diff():
    v1 = VectorFunction([3*t**2, ln(t), 4])
    v2 = VectorFunction([sin(t), t**3, 1-t])
    v3 = VectorFunction([cos(t), sin(t), t])
    sols = [VectorFunction([6*t, 1/t, 0]),
            VectorFunction([cos(t), 3*t**2, -1]),
            VectorFunction([-sin(t), cos(t), 1])]
    for v, s in zip((v1, v2, v3), sols):
        assert v.derive() == s


def test_integration():
    v1 = VectorFunction([(2*t+1)**5, -1/t, 0])
    v2 = VectorFunction([sin(t), cos(t), tan(t)])
    answers = [VectorFunction([16*t**6/3 + 16*t**5 + 20*t**4 + 40*t**3/3 + 5*t**2 + t, -log(t), 0]),
               VectorFunction([-cos(t), sin(t), -ln(cos(t))])]
    for v, a in zip((v1, v2), answers):
        assert v.integrate() == a


def test_domain():
    v1 = VectorFunction([t**2, sqrt(1-t), -1/t])
    v2 = VectorFunction([ln(t+1), 1/(E**t-2), t])
    v3 = VectorFunction([cos(t), sin(t), 5])
    v4 = VectorFunction([ln(t), t+1, E**t])
    v5 = VectorFunction([sin(t), ln(abs(t)), sqrt(4-t)])
    answers = [Union(Interval.Ropen(-oo, 0), Interval.Lopen(0, 1)),
               Union(Interval.open(-1, log(2)), Interval.open(log(2), oo)),
               Reals,
               Interval.open(0, oo),
               Union(Interval.open(-oo, 0), Interval.Lopen(0, 4))]

    for v, d in zip((v1, v2, v3, v4, v5), answers):
        assert v.get_domain() == d


def test_plugin():
    r1, t1 = VectorFunction([-sin(t), cos(t), 1]), pi/2
    r2, t2 = VectorFunction([6*t, 1/t, 0]), 2
    r3, t3 = VectorFunction([cos(t), sin(t), t]), pi/2
    sols = [[-1, 0, 1], [12, 1/2, 0], [0, 1, pi/2]]
    for r, sol in zip(((r1, t1), (r2, t2), (r3, t3)), sols):
        assert r[0].plugin(r[1]) == sol


def test_line_tangent():
    r1, p1 = VectorFunction([ln(t), 2*sqrt(t), t**2]), [0, 2, 1]
    r2, p2 = VectorFunction([sin(t), cos(t), tan(t)]), pi
    l1 = ParametricLine(point=[0, 2, 1], vector=[1, 1, 2])
    l2 = ParametricLine(point=[0, -1, 0], vector=[-1, 0, 1])
    assert r1.get_tangent_line(point=p1) == l1
    assert r2.get_tangent_line(tau=p2) == l2


def test_solve_integration():
    r1, i1, p1 = VectorFunction([E**-t, 3*t**2, 0]), [2, -8, 0], 0
    r2, i2, p2 = VectorFunction([4*t, E**t, 0]), [2, 3, 0], 0
    sols = [VectorFunction([-E**(-t)+3, t**3-8, 0]),
            VectorFunction([2*t**2+2, E**t+2, 0])]
    assert r1.solve_integration(initial=i1, point=p1) == sols[0]
    assert r2.solve_integration(initial=i2, point=p2) == sols[1]



'''
description: vector-value function class
author: ao wang
date: 09/01/2020
'''
from IPython.display import Math, display
from sympy import (E, Intersection, Interval, S, Symbol, cos, diff, integrate,
                   latex, ln, oo, sin, solve, sqrt)
from sympy.abc import t
from sympy.calculus.util import continuous_domain
from sympy.matrices import Matrix
from sympy.vector import CoordSys3D, matrix_to_vector

from axiomathbf.environment import isnotebook

from .parametric_lines import ParametricLine


class VectorFunction():
    '''Vector-Value Function that derives, integrates, and finds the domain of the vector

    :param lst: private list of vectors
    :type lst: lst of ints
    :param vector: matrix object of the vectors
    :type vector: sympy.matrices.dense.MutableDenseMatrix
    '''

    def __init__(self, lst):
        self.__lst = lst
        self.vector = Matrix(lst)

    def __repr__(self):
        if isnotebook():
            display(Math(latex(self.get_vector())))
            return ''
        return self.__str__()

    def __str__(self):
        '''Prints out list'''
        return str(self.__lst).replace('[', '<').replace(']', '>')

    def __eq__(self, other):
        return self.__lst == other.__lst

    def get_vector(self):
        '''Converts Matrix vector into a vector object

        :return: the Vector-Value Function
        :rtype: sympy.vector.vector.VectorAdd
        '''
        C = CoordSys3D('')
        return matrix_to_vector(self.vector, C)

    def derive(self):
        '''Derives each of the functions in the vector function

        :return: the differientation of the Vector-Value Function
        :rtype: axiomathbf.vector_value.VectorFunction
        '''
        return VectorFunction([diff(elem, t) for elem in self.__lst])

    def integrate(self):
        '''Integrates each of the functions in the vector function

        :return: the integrated version of the Vector-Value Function
        :rtype: sympy.vector.vector.VectorAdd
        '''
        return VectorFunction([integrate(elem, t) for elem in self.__lst])

    def get_domain(self):
        '''Returns domain of vector-value function

        :return: the domain of the vector-value function
        :rtype: sympy.sets.sets.Union
        '''

        # Starts domain in all reals
        domain = Interval(-oo, oo)

        # Intersection for each function's domain
        for vector in self.__lst:
            if type(vector) != int:
                domain = Intersection(continuous_domain(
                    vector, t, S.Reals), domain)
        return domain

    def plugin(self, pt):
        '''Returns Vector-Value Function given a point

        :param pt: a point at t
        :type pt: int
        :return: the vector-value function at point t
        :rtype: list of ints
        '''
        return [elem.subs(t, pt) if type(elem) != int else elem for elem in self.__lst]

    def get_tangent_line(self, tau=None, point=None):
        '''Gets the parametric equation of the tan¸gent line to the original function

        :param tau: the time at t
        :type tau: int
        :param point: the point at t
        :type point: list of ints
        :return: the parametic tangent line at t
        :rtype: axiomathbf.vector_value.VectorFunction
        '''
        if tau != None:
            point, vector = self.plugin(tau), self.derive().plugin(tau)
        elif point != None:
            # solves for t with first function
            tau = solve(self.__lst[0], t)[0]
            vector = self.derive().plugin(tau)
        return ParametricLine(point, vector)

    def solve_integration(self, initial, point):
        '''Solve for position function, given a velocity function, point, and start position

        :param point: the point at t
        :type point: list of ints
        :param initial: the initial vector-value position
        :type initial: list of ints
        :return: the position vector-value function
        :rtype: axiomathbf.vector_value.VectorFunction
        '''
        pos = self.integrate()
        plugged = [elem.subs(t, point) for elem in pos.__lst]
        c_lst = [i - p for i, p in zip(initial, plugged)]  # solve for the c's
        return VectorFunction([i+c for i, c in zip(pos.__lst, c_lst)])



import pytest
from axiomathbf import ChainRule
from sympy.abc import x, y, z, w, r, s, t
from sympy import sin, cos, tan, E, pi, ln
import sympy


def test_equation():
    c1 = ChainRule(f=(x, y, z, w), x=(r, s, t), y=(r, t), z=(r, s), w=(s, t))
    c2 = ChainRule(z=(x, y), x=(t,), y=(t,))

    d1, d2, d3 = 'df/dr', 'df/ds', 'df/dt'
    sols = ['∂f/∂x * ∂x/∂r + ∂f/∂y * ∂y/∂r + ∂f/∂z * ∂z/∂r',
            '∂f/∂x * ∂x/∂s + ∂f/∂z * ∂z/∂s + ∂f/∂w * ∂w/∂s',
            '∂f/∂x * ∂x/∂t + ∂f/∂y * ∂y/∂t + ∂f/∂w * ∂w/∂t']
    for d, sol in zip((d1, d2, d3), sols):
        assert c1.get_equation(d) == sol
    assert c2.get_equation('dz/dt') == '∂z/∂x * ∂x/∂t + ∂z/∂y * ∂y/∂t'


def test_solve():
    z1, x1, y1 = 2*x-y, sin(t), 3*t
    z2, x2, y2 = x*sin(y), E**t, pi*t
    z3, x3, y3 = x*y+y**2, t**2, t+1
    z4, x4, y4 = ln(x**2/y), E**t, t**2
    w1, x5, y5, z5 = x**2+y**2+2*z**2, t+1, cos(t), sin(t)
    sols = [2*cos(t)-3, E**t*sin(pi*t)+pi*E**t*cos(pi*t), 3*t**2+4*t+2, 2-2/t]
    for f, sol in zip(((z1, x1, y1), (z2, x2, y2), (z3, x3, y3), (z4, x4, y4)), sols):
        assert ChainRule().solve('dz/dt', z=f[0], x=f[1], y=f[2]).equals(sol)

    assert ChainRule().solve('dw/dt', w=w1, x=x5, y=y5, z=z5).equals(2*t+2+2*sin(t)*cos(t))



'''
description: extrema class
author: ao wang
date: 09/12/2020
'''

import sympy
from IPython.display import Math, display
from sympy import E
from sympy.abc import x, y, z


class Extrema():
    '''The Extrema class calculates the critical points of a function, infer if the points are
    relative minimum, maximum, or saddlepoint. It can also find a function's absolute extremas.

    :param function: the multivariable function
    :type function: sympy.core.add.Add
    '''

    def __init__(self, function):
        self.function = function

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return str(self.function)

    def set_function(self, function):
        '''Sets the function class parameter

        :param function: the multivariable function
        :type function: sympy.core.add.Add
        '''
        self.function = function

    def get_function(self):
        '''Gets the function class parameter

        :return: the function parameter
        :rtype: sympy.core.add.Add
        '''
        return self.function

    def get_critical_points(self):
        '''Finds all critical points of a function

        :return: the critical points
        :rtype: tuples of ints or list of tuples of ints
        '''
        gradient = sympy.derive_by_array(self.function, (x, y))
        stationary_points = sympy.solve(gradient, (x, y))
        if type(stationary_points) == dict:
            return (stationary_points[x], stationary_points[y])
        return stationary_points

    def get_relative(self):
        '''Finds the relative extremas of the function

        :return: determines if point of mini, max, of saddle point
        :rtype: str
        '''
        results = ''
        gradient = sympy.derive_by_array(self.function, (x, y))
        hessian = sympy.Matrix(sympy.derive_by_array(gradient, (x, y)))
        crit_points = sympy.solve(gradient, (x, y))

        if type(crit_points) == dict:
            crit_points = [(crit_points[x], crit_points[y])]
        for point in crit_points:
            hess = hessian.subs({x: point[0], y: point[1]})
            d, fxx = hess.det(), hess[0]
            if d > 0 and fxx > 0:
                results += 'Relative minimum at {}\n'.format(point)
            elif d > 0 and fxx < 0:
                results += 'Relative maximum at {}\n'.format(point)
            elif d < 0:
                results += 'Saddle point at {}\n'.format(point)
            else:
                results += 'Results inconclusive at {}\n'.format(point)
        return results

    def __get_area(self, p1, p2, p3):
        '''Calculates area of triangle given three points

        :param p1: first point
        :type p1: tuple of ints
        :param p2: second point
        :type p2: tuple of ints
        :param p3: third point
        :type p3: tuple of ints
        :return: the area of a triangle
        :rtype: int
        '''
        return abs(p1[0]*(p2[1]-p3[1])+p2[0]*(p3[1]-p1[1])+p3[0]*(p1[1]-p2[1]))

    def __is_inside(self, edge_cases, pt):
        '''Checks to see if a point is inside a triangle

        :param edge_cases: the other points to test out
        :type edge_cases: list of tuples of ints
        :param pt: the point to check
        :type pt: tuple of ints
        :return: whether or not point is in triangle
        :rtype: bool
        '''
        a_total = self.__get_area(edge_cases[0], edge_cases[1], edge_cases[2])
        a1 = self.__get_area(pt, edge_cases[1], edge_cases[2])
        a2 = self.__get_area(edge_cases[0], pt, edge_cases[2])
        a3 = self.__get_area(edge_cases[0], edge_cases[1], pt)
        return a_total == (a1+a2+a3)

    def get_absolute(self, edge_cases):
        '''Calculates the absolute extrema of a function.

        Warning!! Can only check for three edge cases for the closed region (i.e. a triangle border)

        :param edge_cases: the other points to test out
        :type edge_cases: list of tuples of ints
        :return: the absolute extrema
        :rtype: dict of dict of ints
        '''

        # Gets critical points
        points = self.get_critical_points()

        # Get the min and max of the x and y positions from edge cases
        max_y, min_y, max_x, min_x = - \
            float('inf'), float('inf'), -float('inf'), float('inf')
        points = [points]
        for p in edge_cases:
            if p[1] > max_y:
                max_y = p[1]
            if p[1] < min_y:
                min_y = p[1]
            if p[0] > max_x:
                max_x = p[0]
            if p[0] < min_x:
                min_x = p[0]

        if any(points):  # Checks if there are any points to check
            for p in points:
                if p[0] < min_x or p[0] > max_x or p[1] < min_y or p[1] > max_y:
                    points.remove(p)
        else:  # If not, remove the nested empty list
            points.remove([])

        # Loop through possible points for absolute extrema
        for i in range(min_x, max_x+1):
            for j in range(min_y, max_y+1):
                if self.__is_inside(edge_cases=edge_cases, pt=(i, j)):
                    points.append((i, j))

        # Subsitute the values and find the max and min
        maximum, minimum = -float('inf'), float('inf')
        for p in points:
            val = self.function.subs({x: p[0], y: p[1]})
            if val > maximum:
                maximum = val
                max_p = p
            if val < minimum:
                minimum = val
                min_p = p
        return {'max': {'point': max_p, 'value': maximum},
                'min': {'point': min_p, 'value': minimum}}



from IPython import get_ipython


def isnotebook():
    '''Checks to see if the Python environment is in Jupyter Notebook or not.

    Credit: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook/54967911#54967911

    :return: whether or not the working environment is in Jupyter Notebook
    :rtype: bool
    '''
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter



from .chainrule import ChainRule
from .directional_derivative import DirectionalDerivative
from .extrema import Extrema
from .gradient import Gradient
from .mv_function import MVFunction
from .parametric_lines import ParametricLine
from .plane import Plane
from .vector_value import VectorFunction



'''
description: parametric lines in 3D space
author: ao wang
date: 09/02/2020
'''
import math

import sympy
from IPython.display import Math, display
from sympy.matrices import Matrix
from sympy.vector import CoordSys3D, matrix_to_vector

from axiomathbf.environment import isnotebook


class ParametricLine():
    '''A ParametricLine class that determines if lines are parallel,
       intersecting, or skew, point of intersection, and displaying
       the equations in latex.

    :param point: a point
    :type point: sympy.geometry.point.Point3D
    :param vector: a directonal vector
    :type vector: sympy.matrices.dense.MutableDenseMatrix
    '''

    def __init__(self, point, vector):
        self.point = sympy.Point(point)

        # Simplifies the directional vector
        self.vector = Matrix(vector)/sympy.gcd(list(vector))

    def __repr__(self):
        if isnotebook():
            display(Math(self.get_point_vector()))
            return ''
        return self.__str__()

    def __str__(self):
        x, y, z = self.point
        v1, v2, v3 = self.vector
        return '<x, y, z> = <{}, {}, {}> + <{}, {}, {}>t'.format(x, y, z, v1, v2, v3)

    def __eq__(self, other):
        # Checks equality of two Parametric Lines given they have the same point and same direction vectors
        return self.point == other.point and (self.vector == other.vector or self.vector == -other.vector)

    def get_point(self):
        '''Gets the point class attribute

        :return: the point class attribute
        :rtype: sympy.geometry.point.Point3D
        '''
        return self.point

    def set_point(self, point):
        '''Sets the point class attribute

        :param point: the point class attribute
        :type point: sympy.geometry.point.Point3D
        '''
        self.point = point

    def get_vector(self):
        '''Gets the vector class attribute

        :return: the vector class attribute
        :rtype: sympy.matrices.dense.MutableDenseMatrix
        '''
        return self.vector

    def set_vector(self, vector):
        '''Sets the vector class attribute

        :param vector: the vector class attribute
        :type vector: sympy.matrices.dense.MutableDenseMatrix
        '''
        self.vector = vector

    def compare(self, other):
        '''Compares two lines to see if they're intersecting, parallel, or skew

        :param other: the other line
        :type other: axiomathbf.parametric_line.ParametricLine
        :return: string that shows if intersecting, parallel, or skew
        :rtype: str
        '''
        if self.intersect(other):
            if self.vector.dot(other.vector) == 0:
                symbol = 'Perpendicular'
            else:
                symbol = 'Intersecting'
        elif self.vector.cross(other.vector).norm() == 0:
            symbol = 'Parallel'
        else:
            symbol = 'Skew'
        return symbol

    def distance(self, other):
        '''Calculates the distance between another ParametricLine object or sympy.Point object

        :param other: the other 3D object
        :type other: sympy.geometry.point.Point3D or axiomathbf.parametric_line.ParametricLine
        :return: the distance between the two objects
        :rtype: sympy.core.numbers.Floats
        '''
        if isinstance(other, ParametricLine):
            pq = Matrix(other.point-self.point)
            lines = self.compare(other)
            if lines == 'Skew':
                return abs(pq.dot(other.vector.cross(self.vector)))/(other.vector.cross(self.vector).norm())
            elif lines == 'Parallel':
                v = (pq.dot(self.vector)/(self.vector.norm())**2)*self.vector
                return (pq - v).norm()
            else:
                return 0
        elif isinstance(other, sympy.Point):
            pq = Matrix(other - self.point)
            return pq.cross(self.vector).norm()/self.vector.norm()

    def get_point_vector(self):
        '''Returns latex form of the vector in point vector form

        :return: renders latex of parametric line
        :rtype: str
        '''
        x, y, z = self.point
        v1, v2, v3 = self.vector
        return sympy.latex('$\\vec{\\ell(t)} = \\langle' + '{}, {}, {}'.format(x, y, z)
                           + '\\rangle +t \\langle ' + ' {}, {}, {}'.format(v1, v2, v3) + '\\rangle$')

    def intersect(self, other):
        '''Returns the point of intersection of two lines

        :return: the point of intersection
        :rtype: returns list of ints when intersecting, None if not intersecting
        '''
        x, y = sympy.symbols('x y')
        indices = [(i, j) for i in range(3) for j in range(i+1, 3)]

        solutions = []
        for i in indices:
            first, latter = i
            r1 = [self.vector[first], -other.vector[first],
                  other.point[first]-self.point[first]]
            r2 = [self.vector[latter], -other.vector[latter],
                  other.point[latter]-self.point[latter]]
            if r1 == r2:
                continue
            a = Matrix([r1, r2])
            solutions.append(sympy.solve_linear_system(a, x, y))
        if None not in solutions and solutions.count(solutions[0]) == len(solutions):
            return [pt + v*solutions[0][x] for (pt, v) in zip(self.point, self.vector)]
        return None



'''
description: gradient.py
author: Ao Wang
date: 09/18/20
'''

from IPython.display import Math, display
from sympy import Matrix, derive_by_array, latex
from sympy.abc import x, y, z

from axiomathbf.environment import isnotebook


class Gradient():
    '''The Gradient class calculates the gradient vector and the gradient at a point

    :param function: the function for gradient
    :type function: sympy.core.add.Add
    :param vector: the gradient vector
    :type vector: list of int
    :param latex: the latex code for gradient vector
    :type latex: str
    '''

    def __init__(self, function):
        self.function = function
        self.vector = self.__get_gradient()
        self.latex = self.__get_latex

    def __repr__(self):
        if isnotebook():
            display(Math(self.__get_latex()))
            return ''
        return self.__str__()

    def __str__(self):
        v1, v2, v3 = self.vector
        return '<{}, {}, {}>'.format(v1, v2, v3)

    def get_function(self):
        '''Gets the function class attribute

        :return: the function class attribute
        :rtype: sympy.core.add.Add
        '''
        return self.function

    def set_function(self, function):
        '''Sets the function class attribute

        :param function: the function class attribute
        :type function: sympy.geometry.plane.Plane
        '''
        self.function = function

    def __get_gradient(self):
        '''Helper function to get gradient vector

        :return: the differientiation of x, y, z for the function
        :rtype: list of numbers
        '''
        return list(derive_by_array(self.function, (x, y, z)))

    def __get_latex(self):
        '''Formats the latex code for gradient vector

        :return: the latex code for gradient vector
        :rtype: str
        '''
        v1, v2, v3 = self.vector
        return latex('$\\langle{}, {}, {}\\rangle$'.format(v1, v2, v3))

    def at(self, point):
        '''Returns the gradient vector at a certain point

        :param point: a 3D point
        :type point: list of ints
        :return: a gradient at a certain point
        :rtype: sympy.matrices.dense.MutableDenseMatrix
        '''
        p1, p2, p3 = point
        return Matrix([elem.subs(([x, p1], [y, p2], [z, p3]))
                       for elem in self.vector])



from axiomathbf import ParametricLine
import sympy
from math import isclose


def test_compare():
    l1 = ParametricLine(point=[2, 1, 4], vector=[3, -2, 5])
    l2 = ParametricLine(point=[3, -2, -1], vector=[-6, 4, -10])

    l3 = ParametricLine(point=[1, 0, 2], vector=[0, 1, -1])
    l4 = ParametricLine(point=[2, 4, 0], vector=[3, -3, 1])

    l5 = ParametricLine(point=[1, 14, 5], vector=[-2, 1, -1])
    l6 = ParametricLine(point=[0, 4, 3], vector=[1, 3, 1])

    l7 = ParametricLine(point=[2, 4, 1], vector=[5, -1, 1])
    l8 = ParametricLine(point=[3, 1, 0], vector=[6, -1, 1])

    sols = ['Parallel', 'Skew', 'Perpendicular', 'Skew']
    for lines, sol in zip(((l1, l2), (l3, l4), (l5, l6), (l7, l8)), sols):
        assert lines[0].compare(lines[1]) == sol


def test_str():
    l1 = ParametricLine(point=[1, 3, 4], vector=[-3, -2, 3])
    l2 = ParametricLine(point=[4, -2, 0], vector=[-1, 5, -1])
    l3 = ParametricLine(point=[3, 1, 2], vector=[-2, 1, 1])
    sols = ['<x, y, z> = <1, 3, 4> + <-3, -2, 3>t',
            '<x, y, z> = <4, -2, 0> + <-1, 5, -1>t',
            '<x, y, z> = <3, 1, 2> + <-2, 1, 1>t']
    for line, sol in zip((l1, l2, l3), sols):
        assert str(line) == sol


def test_equal():
    l1 = ParametricLine(point=[3, 1, 2], vector=[-2, 1, 1])
    l2 = ParametricLine(point=[3, 1, 2], vector=[4, -2, -2])

    l3 = ParametricLine(point=[1, 3, 4], vector=[-3, -2, 3])
    l4 = ParametricLine(point=[1, 3, 4], vector=[-3, -2, 3])

    l5 = ParametricLine(point=[1, 2, -1], vector=[-1, 3, -2])
    l6 = ParametricLine(point=[3, 1, 0], vector=[4, -1, 1])
    sols = [True, True, False]
    for lines, sol in zip(((l1, l2), (l3, l4), (l5, l6)), sols):
        assert (lines[0] == lines[1]) == sol


def test_distance():
    l1 = ParametricLine(point=[5, 3, 0], vector=[3, 9, 0])
    l2 = ParametricLine(point=[1, 0, 1], vector=[1, 3, 0])

    l3, p1 = ParametricLine(
        point=[2, -4, 0], vector=[3, 1, -2]), sympy.Point([1, -6, 2])

    sols = [sympy.sqrt(91/10), sympy.sqrt(45/14)]
    for i, sol in zip(((l1, l2), (l3, p1)), sols):
        assert isclose(i[0].distance(i[1]), sol) == True


def test_intersect():
    l1 = ParametricLine(point=[1, 3, 5], vector=[2, 5, 2])
    l2 = ParametricLine(point=[0, 11, 4], vector=[1, -1, 1])
    
    l3 = ParametricLine(point=[2, 3, 1], vector=[4, 0, -1])
    l4 = ParametricLine(point=[2, 3, 1], vector=[2, 2, 1])
    sols = [[3, 8, 7], [2, 3, 1]]
    for i, sol in zip(((l1, l2), (l3, l4)), sols):
        assert i[0].intersect(i[1]) == sol



'''
description: chain rule class
author: ao wang
date: 09/12/2020
'''

import sympy
from sympy.abc import x, y, z, w, r, s, t
from IPython.display import display, Math
import re


class ChainRule():
    '''The class is able to compute chain rule based on find the partial derivatives
    given the multivariable functions and also explicitly defined functions.
    
    Here's an example: c1 = ChainRule(f=(x, y, z, w), x=(r, s, t), y=(r, t), z=(r, s), w=(s, t)).

    :param kwargs: the explicit defned functions
    :type kwargs: tuples of sympy.core.symbol.Symbol
    '''

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return str(self.__dict__)

    def __make_diff(self, top, bot):
        '''Helper method that returns the partial derivative symbol

        :param top: the numerator of the partial derivative
        :type top: str
        :param bot: the denominator of the partial derivative
        :return: the string format for partial derivative
        :rtype: str
        '''
        return '∂{}/∂{}'.format(top, bot)

    def get_equation(self, diff):
        '''Calculates the formula for chain rule given the explicit functions 
        and differientiation

        :param diff: the derivative, i.e. dz/dt
        :type diff: str
        :return: the chain rule formula
        :rtype: str
        '''
        diff = diff.split('/')
        root, leaf = diff[0][1], diff[1][1]
        eq = []
        for branch in self.__dict__[root]:
            for var in self.__dict__[str(branch)]:
                if str(var) == leaf:
                    eq.append([self.__make_diff(root, branch),
                               self.__make_diff(branch, var)])
        return ' + '.join([' * '.join(i) for i in eq])

    def get_latex_equation(self, diff):
        '''Calculates the formula for chain rule given the explicit functions 
        and differientiation and formats for LaTeX

        :param diff: the derivative, i.e. dz/dt
        :type diff: str
        :return: the chain rule formula in LaTeX
        :rtype: str
        '''
        match = '∂.\/∂.'
        eq = self.get_equation(diff)
        partials = re.findall(match, eq)
        for p in partials:
            diff = p.split('/')
            root, leaf = diff[0][1], diff[1][1]
            eq = eq.replace(p, ('\\frac{' + '\\partial {}'.format(
                root) + '}{' + '\\partial {}'.format(leaf) + '}'))

        eq = eq.replace(' * ', '')
        return sympy.latex('${}$'.format(eq))

    def solve(self, diff, **kwargs):
        '''Calculates the chain rule given the differientation and sympy functions
        Here's an example: c1 = ChainRule(); print(c1.solve('dz/dt', z=2*x-y, x=sympy.sin(t), y=3*t))

        :param diff: the derivative, i.e. dz/dt
        :type diff: str
        :param kwargs: multivariate functions (sympy.core.add.Add)
        :type kwargs: sympy.core.add.Add
        :return: the chainrule equation
        :rtype: sympy.core.add.Add
        '''
        diff = diff.split('/')
        root, leaf = diff[0][1], diff[1][1]
        top = kwargs[root]
        kwargs.pop(root, None)
        top = top.subs(kwargs)
        return sympy.diff(top, sympy.Symbol(leaf))




from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()
setup(
    name='axiomathbf',         # How you named your package folder (MyLib)
    packages=['axiomathbf'],   # Chose the same as "name"
    version='0.0.4',      # Start with a small number and increase it with every change you make
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    license='MIT',
    # Give a short description about your library
    description='Multivariable Calculus Aid',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ao Wang',                   # Type in your name
    author_email='aw3338@drexel.edu',      # Type in your E-Mail
    # Provide either the link to your github or to your website
    url='https://github.com/AoWangPhilly',
    # Keywords that define your package best
    keywords=['Python', 'Multivariable Calculus', 'Sympy', 'Latex'],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
)



from axiomathbf import Plane, ParametricLine
import sympy
from sympy import sqrt, acos, Point, pi, asin
from sympy.abc import x, y, z
import pytest


def test_str():
    p1, n1 = [1, 2, 1], [-7, -8, -10]
    p2, p3, p4 = [1, 2, 1], [3, -1, 2], [-1, 0, 4]

    plane1 = Plane(p1=p1, normal_vector=n1)
    plane2 = Plane(p1=p2, p2=p3, p3=p4)
    sols = [Plane(eq=-7*(x-1)-8*(y-2)-10*(z-1)),
            Plane(eq=-7*(x-1)-8*(y-2)-10*(z-1))]

    for plane, sol in zip((plane1, plane2), sols):
        assert str(plane) == str(sol)


def test_eq():
    p1 = Plane(eq=3*x-y+2*z+1)
    p2 = Plane(p1=[-3, 2, 5], normal_vector=[3, -1, 2])

    p3 = Plane(p1=[1, 2, 3], normal_vector=[4, -2, 6])
    p4 = Plane(eq=4*(x-1)-2*(y-2)+6*(z-3))

    p5 = Plane(p1=[1, 2, 3], p2=[2, -1, 5], p3=[-1, 3, 3])
    p6 = Plane(eq=-2*(x-1)-4*(y-2)-5*(z-3))

    for p in ((p1, p2), (p3, p4), (p5, p6)):
        assert (p[0] == p[1]) == True


def test_angle():
    p1, p2 = Plane(eq=3*x-2*y+5*z), Plane(eq=-x-y+2*z-3)

    p3, p4 = Plane(eq=2*x-3*y+4*z-5), Plane(eq=3*x+5*y-2*z-7)

    p5, l1 = Plane(
        eq=2*x-y+z-4), ParametricLine(point=[1, 2, -1], vector=[1, -1, 1])
    sols = [acos(9/(sqrt(38*6))), acos(-17/sqrt(1102)), asin(4/sqrt(18))]
    for p, sol in zip(((p1, p2), (p3, p4), (p5, l1)), sols):
        assert p[0].angle(p[1]) == sol


def test_compare():
    p1, p2 = Plane(eq=5*x-3*y+4*z+1), Plane(eq=2*x-2*y-4*z-9)
    p3, p4 = Plane(eq=3*x-2*y+z+3), Plane(eq=5*x+y-6*z-8)
    p5, p6 = Plane(eq=3*x-2*y+z+3), Plane(eq=-6*x+4*y-2*z-1)

    p7 = Plane(eq=5*x-3*y+4*z+1)
    l1 = ParametricLine(point=[2, 3, 5], vector=[2, -2, -4])
    p8 = Plane(eq=5*x-3*y+4*z+1)
    l2 = ParametricLine(point=[2, 3, 5], vector=[5/2, -3/2, 2])
    sols = ['Perpendicular', 'Neither parallel nor perpendicular',
            'Parallel', 'Parallel', 'Perpendicular']
    for t, sol in zip(((p1, p2), (p3, p4), (p5, p6), (p7, l1), (p8, l2)), sols):
        assert t[0].compare(t[1]) == sol


def test_distance():
    plane1, p1 = Plane(eq=x+2*y+3*z-5), Point([2, -1, 4])
    plane2, plane3 = Plane(eq=2*x-4*y+5*z+2), Plane(eq=x-2*y+(5/2)*z-5)
    plane4, line1 = Plane(
        eq=2*x+3*y+4*z-12), ParametricLine(point=[2, 3, 3], vector=[-1, 2, 4])
    sols = [7/sympy.sqrt(14), 12/sympy.sqrt(45), 13/sqrt(29)]
    for p, sol in zip(((plane1, p1), (plane2, plane3), (plane4, line1)), sols):
        assert p[0].distance(p[1]) == sol


def test_intersect():
    p1, p2 = Plane(eq=x+y+z-7), Plane(eq=2*x+4*z-6)
    p3, p4 = Plane(eq=x+y+z+1), Plane(eq=x+2*y+3*z+4)
    sols = [ParametricLine(point=[3, 4, 0], vector=[-2, 1, 1]),
            ParametricLine(point=[2, -3, 0], vector=[1, -2, 1])]
    for p, sol in zip(((p1, p2), (p3, p4)), sols):
        assert p[0].intersect(p[1]) == sol



from axiomathbf import MVFunction, Gradient, DirectionalDerivative, Plane, ParametricLine
import pytest
from sympy.abc import x, y, z
from sympy import sin, ln, E, pi, cos, tan, atan, sqrt, Matrix, asin, simplify

# Gradient


def test_str():
    f1 = Gradient(x**3-2*x*y**2)
    f2 = Gradient(3*x**2*y**3)
    sols = ['<3*x**2 - 2*y**2, -4*x*y, 0>',
            '<6*x*y**3, 9*x**2*y**2, 0>']
    for f, sol in zip((f1, f2), sols):
        assert str(f) == sol


def test_gradient_vector():
    f1 = Gradient(x**3-2*x*y**2)
    f2 = Gradient(3*x**2*y**3)
    sols = [[3*x**2-2*y**2, -4*x*y, 0],
            [6*x*y**3, 9*x**2*y**2, 0]]
    for f, sol in zip((f1, f2), sols):
        assert f.vector == sol


def test_at():
    f1, p1 = Gradient(3*x*y-y**2*x**3), [1, -1, 0]
    f2, p2 = Gradient(cos(2*x-y**2)), [pi/4, 0, 0]
    f3, p3 = Gradient(4*x*y*z-y**2*z**3+4*z**3*y), [2, 3, 1]
    sols = [Matrix([-6, 5, 0]), Matrix([-2, 0, 0]), Matrix([12, 6, 33])]
    for f, sol in zip(((f1, p1), (f2, p2), (f3, p3)), sols):
        assert f[0].at(f[1]) == sol

# DirectionalDerivative


def test_directional_diff():
    f1 = DirectionalDerivative(
        function=x**4-y**4, point=[0, -2, 0], unit_vector=[sqrt(2)/2, sqrt(2)/2, 0])
    f2 = DirectionalDerivative(
        function=y*sin(x), point=[pi/2, 1, 0], unit_vector=[1, -1, 0])
    f3 = DirectionalDerivative(
        function=E**x*cos(y*z), point=[1, pi, 0], unit_vector=[-2, 1, -3])
    sols = [32/sqrt(2), -1/sqrt(2), -2*E/(sqrt(14))]
    for f, sol in zip((f1, f2, f3), sols):
        assert f.value == sol


def test_info():
    f1 = DirectionalDerivative(E**(x*y**2), [1, 3, 0])
    f2 = DirectionalDerivative(sqrt(4-x**2-y**2-z**2), [1, -1, 0])
    f3 = DirectionalDerivative(x**3*y*z**2, [2, -1, 1])
    sols = [(3*sqrt(13)*E**(9), [3*sqrt(13)/13, 2*sqrt(13)/13, 0]),
            (1, [-sqrt(2)/2, sqrt(2)/2, 0])]
    for f, sol in zip((f1,f2), sols):
        assert f.info() == sol
    assert f3.info(increasing=False) == (-4*sqrt(29), [3*sqrt(29)/29, -2*sqrt(29)/29, 4*sqrt(29)/29])

# MVFunction


def test_linearization():
    f1 = MVFunction(x**2-y**2, [1, 2, 0])
    f2 = MVFunction((x+y)/(x-y), [2, 1, 0])
    f3 = MVFunction(E**x*sin(y), [ln(3), pi/2, 0])
    f4 = MVFunction(ln(x**2-y**2), [2, sqrt(3), 0])
    f5 = MVFunction(atan(x/y), [1, 1, 0])
    sols = [3+2*x-4*y, 3-2*x+4*y, 3-3*ln(3)+3*x,
            -2+4*x-2*sqrt(3)*y, pi/4+(1/2)*x-(1/2)*y]
    for f, sol in zip((f1, f2, f3, f4, f5), sols):
        assert f.get_linearization().equals(sol)


s1 = MVFunction(ln(x+y+z), [-1, E**2, 1])
s2 = MVFunction(x**2+y**2+z**2-1, [sqrt(3)/3, sqrt(3)/3, sqrt(3)/3])
s3 = MVFunction(x**2-x*y+z**2, [2, 2, 3])


def test_tangent_plane():
    sols = [Plane(eq=x+y+z-E**2), Plane(eq=x+y+z-sqrt(3)), Plane(eq=x-y+3*z-9)]
    for s, sol in zip((s1, s2, s3), sols):
        assert s.get_tangent_plane() == sol


def test_normal_line():
    sols = [ParametricLine(point=[-1, E**2, 1], vector=[1, 1, 1]),
            ParametricLine(point=[sqrt(3)/3, sqrt(3)/3,
                                  sqrt(3)/3], vector=[1, 1, 1]),
            ParametricLine(point=[2, 2, 3], vector=[1, -1, 3])]
    for s, sol in zip((s1, s2, s3), sols):
        assert s.get_normal_line() == sol
